#ifndef MMASS_H
#define MMASS_H
#include <pthread.h>
#include "MMASSDefs.h"
#include "MASS.h"
#include <string>
#include <map>
#include <stdexcept>

namespace MMASS {
	class CMMASS {
	public:
		class logic_error : public std::logic_error {
		public:
			logic_error(const string& msg = "") : std::logic_error(msg) {}
		};

		class invalid_argument : public std::invalid_argument {
		public:
			invalid_argument(const string& msg = "") : std::invalid_argument(msg) {}
		};

	protected:
		spMASS_map masss;
		// interfaces

	public:
		unsigned long LIMIT;
		~CMMASS();

		static const spMMASS getSingleton();

		const spMMASS getSelf() const;

		spMASS createMASS(const string &name_) throw (logic_error, invalid_argument);
		spMASS getMASS(const string &name_) throw (logic_error, invalid_argument);
		const bool hasMASS() const;
		const bool hasMASS(const string &name_) const;
		spMASS_map::const_iterator beginMASSs() const;
		spMASS_map::const_iterator endMASSs() const;
		const spMASS getUnfinishedMASS() const;
		const size_t numOfMASSs() const;

		void finish()  throw (logic_error);
		const bool isReady() const;

		volatile unsigned long getTurn();

		volatile bool finishThread();
		void setFinishThread(bool b);

		std::map<const std::string, pthread_t>::iterator beginAgentThreads();
		std::map<const std::string, pthread_t>::iterator endAgentThreads();
		volatile pthread_t& getEnvironmentThread();

		void initiate(pthread_mutex_t *mutex_, pthread_cond_t *cond_) throw (logic_error);

	private:
		static spMMASS           self;
		bool                     ready;
		volatile unsigned long   turn;
		volatile bool            finish_thread;
		volatile pthread_mutex_t mutex;
		volatile pthread_cond_t  cond;

		std::map<const std::string, pthread_t> threads;
		pthread_t envThread;

		void nextTurn();

		volatile pthread_mutex_t& getMutex();
		volatile pthread_cond_t& getCond();

		bool act(spAgent agent, unsigned long turn_);

		CMMASS();

		friend class Agent; // Temporary
		friend class Site;
		friend void* _environmentBehaviourThread(void *arg);
	};

	void _clearOccupiedSites(spMMASS mmass);
	bool _transportManagement(spAgent agent, spAction action, unsigned long turn);
	bool _emitManagement(spAgent agent, spAction action, unsigned long turn);
	bool _reactionManagement(spAgent agent, spAction action, unsigned long turn);
	bool _triggerManagement(spAgent agent, spAction action);
	void _clearFields(spMMASS mmass);
	void _performActions(spAgent_set &agents);
	void printInfo(spAgent_set &agents); // TEMP!!!!!!!!!!!
	void printSitesState(spMMASS mmass); // TEMP!!!!!!!!!!!
}

#endif