#include "Agent.h"
#include "Action.h"
#include "Site.h"
#include "Space.h"
#include "MASS.h"
#include "MMASS.h"
using namespace MMASS;

Agent::~Agent() {
	int rc;
	rc = pthread_mutex_destroy(&const_cast<pthread_mutex_t>(mutex));
	rc += pthread_cond_destroy(&const_cast<pthread_cond_t>(cond));
	if (rc != 0) printf("Agent::~Agent() ln11\n");
}

const string& Agent::getName() const {
	return name;
}

const spAgent Agent::getSelf() const {
	return self.lock();
}

const bool Agent::hasSite() const {
	return site.lock();
}

spSite Agent::getSite() const {
	return site.lock();
}

void Agent::setSite(spSite site_) throw (logic_error, invalid_argument) {
	if (!site_)
		throw invalid_argument("Agent::setSite(spSite site_) - "
			"site_: null\n"
			"Can't set an agent's site to null. (agent: \""
			+name+"\".\n");

	if (site_->getSpace()->getMASS() != mass.lock())
		throw logic_error("Agent::setSite(spSite site_) - "
			"site_: "+site_->getName()+"\n"
			"This site's MASS (\""+site_->getSpace()->getMASS()->getName()+
			"\") isn't the same agent's (\""+name+"\") MASS (\""
			+mass.lock()->getName()+"\").\n");

	spAgent a = site_->getAgent();
	if (a) {
		if (a == self.lock()) return;

		throw invalid_argument("Agent::setSite(spSite site_) - "
			"site_: "+site_->getName()+"\n"
			"This site already has the agent \""+a->getName()+"\". "
			"Can't set the agent \""+name+"\"'s site.\n");
	}

	spSite s = site.lock();
	if (s) s->setAgentToNull();

	site = site_;
	try {
		site_->setAgent(self.lock());
	} catch (Site::logic_error &e) {
		throw invalid_argument(string("Agent::setSite(spSite site_) -\n"
			"<-- ")+e.what());
	}
}

const spMASS Agent::getMASS() const {
	return mass.lock();
}

Agent::Agent(const string &name_) : name(name_) {
	int rc;
	turn = 0;
	action_counter = 0;
	action_denied_counter = 0;
	action_accepted_counter = 0;
	finish_thread = false;
	action_accepted = false;
	action_performed = false;
	rc = pthread_mutex_init(&const_cast<pthread_mutex_t>(mutex), NULL);
	rc += pthread_cond_init(&const_cast<pthread_cond_t>(cond), NULL);
	if (rc != 0) printf("Agent::Agent(const string &name_) ln83\n");
}

volatile pthread_mutex_t& Agent::getMutex() {
	return mutex;
}

volatile pthread_cond_t& Agent::getCond() {
	return cond;
}

void Agent::setAction(spAction action_) {
	this->action = action_;
}

spAction Agent::getPossibleAction() {
	return possibleAction;
}

void Agent::setPossibleAction(spAction action) {
	suspendedAction.reset();
	suspendedAction = possibleAction;
	possibleAction.reset();
	possibleAction = action;
}

spActionInfo_set::const_iterator Agent::beginDeniedActions() {
	return deniedActions.begin();
}

volatile unsigned long Agent::getCoefficient(const string &fieldIdentifier) {
	if (fieldsCoefficients.count(fieldIdentifier) == 0) return 0;
	return fieldsCoefficients[fieldIdentifier];
}

const double Agent::getThreshold(const string &fieldIdentifier) {
	if (fieldsThresholds.count(fieldIdentifier) == 0) return 0;
	return fieldsThresholds[fieldIdentifier];
}

spActionInfo_set::const_iterator Agent::endDeniedActions() {
	return deniedActions.end();
}

void Agent::setCoefficient(const string &fieldIdentifier, unsigned long coefficient) {
	fieldsCoefficients[fieldIdentifier] = coefficient;
}

void Agent::setThreshold(const string &fieldIdentifier, const double threshold) {
	fieldsThresholds[fieldIdentifier] = threshold;
}

volatile unsigned long Agent::getTurn() {
	return turn;
}

// TODO: Create state automaticaly!!!!!
spAgent Agent::createAgent(Agent::AgentFactory &factory, const std::string &name_) {
	Agent *agent = factory.create(name_);
	spAgent agent_(agent);
	agent_->setSelf(agent_);

	return agent_;
}

void Agent::setSelf(wpAgent self_) {
	self = self_;
}

void Agent::setMASS(const wpMASS mass_) {
	mass = mass_;
}

void Agent::setSiteToNull() {
	spSite s = site.lock();
	if (!s) return;
	site.reset();
	s->setAgentToNull();
}

spAction Agent::collectAction(unsigned long turn_) {
	return this->possibleAction;
}

spAction Agent::getAction() {
	return action;
}

void Agent::setActionToNull() {
	action.reset();
	possibleAction.reset();
	suspendedAction.reset();
}

void Agent::addDeniedAction(spActionInfo action) {
	deniedActions.insert(action);
}

bool Agent::hasDeniedAction(const string &actions_identifier) {
	bool reactionDenied = false;

	spActionInfo_set::const_iterator action_i, a_end = this->deniedActions.end();
	for (action_i = this->deniedActions.begin(); action_i != a_end; ++action_i)
		if ((*action_i)->getIdentifier() == actions_identifier)
			reactionDenied = true;

	return reactionDenied;
}

void Agent::clearDeniedActions() {
	deniedActions.clear();
}

void Agent::setActionAccepted(bool b) {
	action_accepted = b;
}

volatile bool Agent::actionAccepted() {
	return action_accepted;
}

volatile bool Agent::actionPerformed() {
	return this->action_performed;
}

void Agent::setActionPerformed(bool b) {
	this->action_performed = b;
}

volatile bool Agent::finishThread() {
	return this->finish_thread;
}

void Agent::setFinishThread(bool b) {
	this->finish_thread = b;
}

void Agent::nextTurn() {
	turn++;
}

spAgentInfo Agent::createAgentInfo() {
	return AgentInfo::createAgentInfo(this->name, this->getTypeName(), this->site.lock()->getName());
}

void Agent::performAction() {
	action_accepted_counter++; // TEMPORARY
	if (!this->action) action_counter++; // TEMPORARY
	this->action.reset();
	this->action = this->possibleAction;
	this->possibleAction.reset();
	//printf("<font color=\"#006600\">MMASS  - %lu - performReact - reaction performed for \"%s\"</font><br>\n", this->getSite()->getSpace()->getMASS()->getMMASS()->getTurn(), this->getName().c_str()); fflush(NULL); // este d para tirar
}

void Agent::notifyFailure(unsigned long turn) {
	action_denied_counter++; // TEMPORARY
	this->possibleAction.reset();
	this->possibleAction = this->suspendedAction;
	this->suspendedAction.reset();
	//printf("<font color=\"#006600\">MMASS  - %lu - notifyFailure - action denied for \"%s\" (<font color=\"#0000ff\">signal</font>)</font><br>\n", this->getSite()->getSpace()->getMASS()->getMMASS()->getTurn(), this->getName().c_str()); fflush(NULL); // este d para tirar
}

void* MMASS::_agentBehaviourThread(void *arg) { 
	spAgent agent = *((spAgent*)arg);
	unsigned long turn = 0;
	spContext nullContext;
	spAction nullAction;

	pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	while (!agent->finishThread()) {
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));

		spContext localContext = agent->getSite()->sense(agent, turn);
		pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
		if ((!agent->actionPerformed()) && (agent->getPossibleAction() == nullAction) && (agent->getAction() == nullAction) && (localContext != nullContext))
			agent->actionSelect(localContext);
		localContext.reset();
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));
		localContext.reset();
		bool outcome = agent->getSite()->act(agent, turn);

		pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
		if (outcome != false) {
			agent->clearDeniedActions();
			turn++;
		}
		//printf("<font color=\"#000099\">\"%s\" - %lu</font><br>\n",  agent->getName().c_str(), agent->getTurn()); fflush(NULL); // este d para tirar
	}
	pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));

	//agent->printState();
	agent.reset();
	pthread_exit(NULL);
	return NULL;
}

const string& AgentInfo::getAgentName() {
	return agentName;
}

const string& AgentInfo::getTypeName() {
	return typeName;
}

const string& AgentInfo::getSiteName() {
	return siteName;
}

AgentInfo::AgentInfo(const string &agentName_, const string &typeName_, const string &siteName_) : agentName(agentName_), typeName(typeName_), siteName(siteName_) { }

spAgentInfo AgentInfo::createAgentInfo(const string &agentName_, const string &typeName_, const string &siteName_) {
	return spAgentInfo(new AgentInfo(agentName_, typeName_, siteName_));
}