#ifndef ACTION_H
#define ACTION_H

#ifdef MMASSDLL_EXPORTS
#define MMASSDLL_API __declspec(dllexport)
#else
#define MMASSDLL_API __declspec(dllimport)
#endif

#include "MMASSDefs.h"
#include "Agent.h"
#include <OgreNoMemoryMacros.h>
#include <set>
#include <string>

namespace MMASS {
	using namespace std;

	class MMASSDLL_API Action {
	public:
		virtual ~Action() = 0;
		spAgentInfo_set& getReactionPartners();
		spAgentInfo_set::const_iterator beginReactionPartners();
		spAgentInfo_set::const_iterator endReactionPartners();

		const string& getDestinationSiteName();

		spAction getSelf();

		virtual void performAction(spAgent agent) = 0;

	protected:
		spAgentInfo_set reactionPartners;
		const string    siteName;
		spField         field;

		Action();
		Action(spField field_);
		Action(const string &siteName_);
		Action(spAgentInfo_set &names);

		void setSelf(wpAction self_);

		virtual const string getIdentifier() = 0;

	private:
		wpAction self;

		virtual bool manage(spAgent agent, spAction action) = 0;
		virtual spActionInfo createActionInfo() = 0;

		/* Hiding methods. */
		Action(const Action&);
		Action& operator=(const Action&);

		friend class Reaction;
		friend MMASSDLL_API bool _manage(spAgent agent, spAction action);
		friend MMASSDLL_API bool _reactionManagement(spAgent agent, spAction action);
		friend MMASSDLL_API bool _triggerManagement(spAgent agent, spAction action);
		friend MMASSDLL_API bool _transportManagement(spAgent agent, spAction action);
	};

	class MMASSDLL_API ActionInfo {
	public:
		const string& getIdentifier();
		spAgentInfo_set::const_iterator beginReactionPartners();
		spAgentInfo_set::const_iterator endReactionPartners();

	protected:
		const string    actionIdentifier;
		spAgentInfo_set reactionPartners;
		const string    siteName;
		spField         field;

	private:
		ActionInfo(const string &actionIdentifier_);
		ActionInfo(const string &actionIdentifier_, spField field_);
		ActionInfo(const string &actionIdentifies_, const string &siteName_);
		ActionInfo(const string &actionIdentifier_, spAgentInfo_set &reactionPartners_);
		static spActionInfo createActionInfo(const string &actionIdentifier_);
		static spActionInfo createActionInfo(const string &actionIdentifier_, spField field_);
		static spActionInfo createActionInfo(const string &actionIdentifier_, const string &siteName_);
		static spActionInfo createActionInfo(const string &actionIdentifier_, spAgentInfo_set &reactionPartners_);

		ActionInfo(const ActionInfo&);
		ActionInfo& operator=(const ActionInfo&);

		friend class Action;
		friend class Reaction;
		friend class Trigger;
		friend class Transport;
		friend class Emit;
	};

	class MMASSDLL_API Reaction : public Action {
	protected:
		Reaction(spAgentInfo_set &names);

	private:
		bool manage(spAgent agent, spAction action);
		spActionInfo createActionInfo();

		Reaction(const Reaction&);
		Reaction& operator=(const Reaction&);

		friend MMASSDLL_API bool _reactionManagement(spAgent agent, spAction action);
	};

	template<typename AgentClass, typename StateClass>
	class CReaction : public Reaction {
	public:
		void performAction(spAgent agent) {
			this->act(agent, AgentClass::extractState(agent));
		}
	protected:
		CReaction(spAgentInfo_set &names) : Reaction(names) { }
		virtual void act(spAgent agent, boost::shared_ptr<StateClass> state) = 0;
	};

	class MMASSDLL_API Transport : public Action {
	public:
		~Transport();

		static spAction createTransport(const string &siteName_);

		void performAction(spAgent agent);

	protected:
		Transport(const string &siteName_);

		const string getIdentifier() { return "Transport"; }

		spActionInfo createActionInfo();

	private:
		bool manage(spAgent agent, spAction action);

		friend MMASSDLL_API bool _transportManagement(spAgent agent, spAction action);
		friend MMASSDLL_API bool _reactionManagement(spAgent agent, spAction action);
	};

	class MMASSDLL_API Trigger : public Action {
	public:
		static spAction createTrigger() { return spAction(new Trigger()); }

		void performAction(spAgent agent) { }

	protected:
		const string getIdentifier() { return "Trigger"; } // TEMPORARY!!!

	private:
		Trigger() { }

		bool manage(spAgent agent, spAction action) { return _triggerManagement(agent, action); }
		spActionInfo createActionInfo() { return ActionInfo::createActionInfo(this->getIdentifier()); }

		friend MMASSDLL_API bool _triggerManagement(spAgent agent, spAction action);
	};

	class MMASSDLL_API Emit : public Action {
	public:
		~Emit();

		static spAction createEmit(spField field);

		void performAction(spAgent agent);

	protected:
		const string getIdentifier() { return "Emit"; }

	private:
		Emit(spField field_);

		bool manage(spAgent agent, spAction action);
		spActionInfo createActionInfo();

		friend MMASSDLL_API bool _emitManagement(spAgent agent, spAction action);
	};
}

#endif