#include "stdafx.h"
#include "MMASS.h"
#include "Action.h"
#include "Site.h"
#include "Space.h"

#include "Field.h" // TEMPORARY

#include <map>
using namespace MMASS;

CMMASS::~CMMASS() {
	int rc;
	pthread_mutex_destroy(&const_cast<pthread_mutex_t>(mutex));
	pthread_cond_destroy(&const_cast<pthread_cond_t>(cond));
}

const spMMASS CMMASS::getSingleton() {
	return self;
}

const spMMASS CMMASS::getSelf() const {
	return self;
}

spAgent_set CMMASS::getAgentSet() {
	spAgent_set agents;

	spMASS_map::const_iterator mass_i, m_end = this->masss.end();
	for (mass_i = this->masss.begin(); mass_i != m_end; ++mass_i) {
		spMASS mass = (*mass_i).second;

		spAgent_map::const_iterator agent_i, a_end = mass->endAgents();
		for (agent_i = mass->beginAgents(); agent_i != a_end; ++agent_i)
			agents.insert((*agent_i).second);
	}

	return agents;
}

spMASS CMMASS::createMASS(const string &name_) throw (logic_error, invalid_argument) {
	if (ready)
		throw logic_error("CMMASS::createMASS(const string &name_)"
			" - name_: "+name_+"\n"
			"The MMASS is ready. Can't create MASS with a finished MMASS.\n");

	if (this->hasMASS(name_)) {
		throw invalid_argument("CMMASS::createMASS(const string &name_)"
			" - name_: "+name_+"\n"
			"A MASS with this name already exists.\n");
	}

	spMASS mass = MASS::createMASS(name_);
	
	mass->setMMASS(self);
	masss[name_] = mass;
	return mass;
}

spMASS CMMASS::getMASS(const string &name_) throw (logic_error, invalid_argument) {
	if (!this->hasMASS()) {
		throw logic_error(string("CMMASS::getMASS(const string &name_)"
			" - name_: "+name_+"\n"
			"There is no MASS. Create one first before "
			"getting it.\n"));
	}

	if (!this->hasMASS(name_)) {
		throw invalid_argument(string("CMMASS::getMASS(const string &name_)"
			" - name_: "+name_+"\n"
			"There is no MASS with this name.\n"));
	}

	return masss[name_];
}

const bool CMMASS::hasMASS() const {
	return (!masss.empty());
}

const bool CMMASS::hasMASS(const string &name_) const {
	return (masss.count(name_) > 0);
}

spMASS_map::const_iterator CMMASS::beginMASSs() const {
	return masss.begin();
}

spMASS_map::const_iterator CMMASS::endMASSs() const {
	return masss.end();
}

const spMASS CMMASS::getUnfinishedMASS() const {
	spMASS_map::const_iterator mass_i, end = masss.end();
	for (mass_i = masss.begin(); mass_i != end; ++mass_i)
		if (!(*mass_i).second->isReady())
			return (*mass_i).second;

	return spMASS();
}

const size_t CMMASS::numOfMASSs() const {
	return masss.size();
}

void CMMASS::finish() throw (logic_error) {
	spMASS unfinished = this->getUnfinishedMASS();
	if (unfinished)
		throw logic_error("CMMASS::finish() - the MASS \""
			+unfinished->getName()+"\" is not finished.\n");

	ready = true;
}

const bool CMMASS::isReady() const {
	return ready;
}

void CMMASS::initiate()  throw (logic_error) {
	if (!this->isReady())
		throw logic_error(string("CMMASS::initiate() -\n"
			"The MMASS isn't ready yet.\n"));

	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	pthread_create(&this->envThread, &attr, _environmentBehaviourThread, (void*)NULL);

	// _environmentBehaviourThread(NULL);

	/*
	spAgent_set agents;
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	spMASS_map::const_iterator mass_i, m_end = this->endMASSs();
	for (mass_i = this->beginMASSs(); mass_i != m_end; ++mass_i) {
		spMASS mass = (*mass_i).second;

		spAgent_map::const_iterator agent_i, a_end = mass->endAgents();
		for (agent_i = mass->beginAgents(); agent_i != a_end; ++agent_i) {
			pthread_create(&threads[(*agent_i).second->getName()], &attr,
				_agentBehaviourThread, (void*)&((*agent_i).second));
			agents.insert((*agent_i).second);
		}
	}

	ThreadArgument *argument = new ThreadArgument(agents);
	pthread_create(&this->envThread, &attr, _environmentBehaviourThread, (void*)argument);
	// _environmentBehaviourThread((void*) argument);

	pthread_attr_destroy(&attr);
	agents.clear();
	*/

	/* ESPERA PELAS THREADS TERMINAR, VOU REITRAR MAIS TARDE!!!!! <-----------*/
	/*
	std::map<const std::string, pthread_t>::const_iterator thread_i,
		t_end = threads.end();
	for (thread_i = threads.begin(); thread_i != t_end; ++thread_i)
		pthread_join((*thread_i).second, NULL);
	pthread_join(envThread, NULL);
	*/
}

volatile unsigned long CMMASS::getTurn() {
	return turn;
}

volatile unsigned long CMMASS::getOgreTurn() {
	return ogreTurn;
}

volatile pthread_mutex_t& CMMASS::getMutex() {
	return mutex;
}

volatile pthread_cond_t& CMMASS::getCond() {
	return cond;
}

volatile pthread_t& CMMASS::getEnvThread() {
	return envThread;
}

volatile bool CMMASS::finishThread() {
	return finish_thread;
}

void CMMASS::setFinishThread(volatile bool b) {
	finish_thread = b;
}

bool CMMASS::waitOgre() {
	return wait_ogre;
}

void CMMASS::setWaitOgre(bool b) {
	wait_ogre = b;
}

void CMMASS::nextTurn() {
	turn++;
}

void CMMASS::nextOgreTurn() {
	ogreTurn++;
}

/*
bool CMMASS::act(spAgent agent, unsigned long turn_) {
	bool ret = false;
	spAction nullAction;

	pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	while ((!agent->finishThread()) && (agent->getPossibleAction() != nullAction) && (agent->getAction() == nullAction) && (!agent->actionPerformed()) && (this->turn <= turn_)) { // este d pra tirar!
		//printf("<font color=\"#000099\">\"%s\" - %lu - act - <font color=\"#0000ff\">waiting</font></font><br>\n",  agent->getName().c_str(), turn_); fflush(NULL); // este d pra tirar!
		pthread_cond_wait(&const_cast<pthread_cond_t>(agent->getCond()), &const_cast<pthread_mutex_t>(agent->getMutex()));
		//printf("<font color=\"#000099\">\"%s\" - %lu - act - <font color=\"#0000ff\">signaled</font></font><br>\n",  agent->getName().c_str(), turn_); fflush(NULL); // este d pra tirar!
	}

	if ((agent->getAction() != nullAction) || (this->turn > turn_) || (agent->actionAccepted())) { // este d pra tirar!
		//printf("<font color=\"#000099\">\"%s\" - %lu - act - action was accepeted</font><br>\n",  agent->getName().c_str(), turn_); fflush(NULL); // este d pra tirar!
		agent->getPossibleAction().reset();
		ret = true;
	}

	agent->setActionPerformed(false);
	//printf("<font color=\"#000099\">\"%s\" - %lu - act - action performed = false</font><br>\n",  agent->getName().c_str(), turn_); fflush(NULL); // este d pra tirar!
	pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	

	return ret;
}
*/

CMMASS::CMMASS() {
	LIMIT = 1000;
	turn = 0;
	ogreTurn = 0;
	ready = false;
	finish_thread = false;
	wait_ogre = false;
	pthread_mutex_init(&const_cast<pthread_mutex_t>(mutex), NULL);
	pthread_cond_init(&const_cast<pthread_cond_t>(cond), NULL);
}

MMASSDLL_API void MMASS::_clearOccupiedSites(spMMASS mmass) {
	spMASS_map::const_iterator mass_i, m_end = mmass->endMASSs();
	for (mass_i = mmass->beginMASSs(); mass_i != m_end; ++mass_i)
		(*mass_i).second->clearOccupiedSites();
}

MMASSDLL_API void MMASS::_clearFields(spMMASS mmass) {
	spMASS_map::const_iterator mass_i, m_end = mmass->endMASSs();
	for (mass_i = mmass->beginMASSs(); mass_i != m_end; ++mass_i) {
		spSpace space = (*mass_i).second->getSpace();
		spSite_map::const_iterator site_i, s_end = space->endSites();
		for (site_i = space->beginSites(); site_i != s_end; ++site_i) {
			(*site_i).second->clearFields();
		}
	}
}


MMASSDLL_API void MMASS::_performActions(spAgent_set &agents) {
	spAgent_set::iterator agent_i, a_end = agents.end();
	for (agent_i = agents.begin(); agent_i != a_end; ++agent_i) {
		spAction action = (*agent_i)->getAction();
		action->performAction((*agent_i));
		(*agent_i)->setActionToNull();
		(*agent_i)->nextTurn();
		(*agent_i)->setActionPerformed(false);
		(*agent_i)->setActionAccepted(false);
		(*agent_i)->clearDeniedActions();
	}
}


MMASSDLL_API bool MMASS::_triggerManagement(spAgent agent, spAction action) {
	agent->performAction();
	agent->setActionAccepted(true);
	agent->setActionPerformed(true);
	return true;
}

MMASSDLL_API bool MMASS::_transportManagement(spAgent agent, spAction action) {
	spMASS mass = agent->getMASS();
	set<const string> occupiedSites = mass->getOccupiedSites();
	const string currentSite = agent->getSite()->getName();
	const string destinationSite = action->getDestinationSiteName();

	set<const string>::iterator site = occupiedSites.find(destinationSite);
	if (site != occupiedSites.end()) {
		agent->addDeniedAction(action->createActionInfo());
		//printf("<font color=\"#006600\">MMASS  - %lu - _transportManagement - <ins>action denied</ins> for \"%s\", action identifier is %s</font><br>\n", turn, agent->getName().c_str(), action->createActionInfo()->getIdentifier().c_str()); fflush(NULL);
		agent->notifyFailure();
		agent->setActionAccepted(false);
		agent->setActionPerformed(true);
		return false;
	}

	agent->performAction();
	mass->eraseOccupiedSite(currentSite); // Maybe it can be removed (1)
	mass->addOccupiedSite(destinationSite);
	agent->setActionAccepted(true);
	agent->setActionPerformed(true);
	return true;
}

MMASSDLL_API bool MMASS::_emitManagement(spAgent agent, spAction action) {
	agent->performAction();
	agent->setActionAccepted(true);
	agent->setActionPerformed(true);
	return true;
}

MMASSDLL_API bool MMASS::_reactionManagement(spAgent agent, spAction action) {
	spMASS mass = agent->getSite()->getSpace()->getMASS();
	bool agreed = true;
	spAgentInfo_set involvedAgents = action->getReactionPartners();
	spAgent_set reactingAgents;
	reactingAgents.insert(agent);

	spAgentInfo_set::iterator info_i, i_end = involvedAgents.end();
	for (info_i = involvedAgents.begin(); info_i != i_end; ++info_i) {
		spAgent agent = mass->getAgent((*info_i)->getAgentName());
		if (!agent->agreeReaction(involvedAgents)) {
			agreed = false;
			break;
		}

		reactingAgents.insert(agent);
	}

	spAgent_set::iterator agent_i, a_end = reactingAgents.end();
	if (agreed) {
		spAction nullAction;
		for (agent_i = reactingAgents.begin(); agent_i != a_end; ++agent_i) {
			if ((*agent_i)->getAction() == nullAction)
				(*agent_i)->setActionPerformed(true);
			else if ((*agent_i)->getAction()->getIdentifier() == "Transport") {
				const string currentSite = (*agent_i)->getSite()->getName();
				const string destinationSite = (*agent_i)->getAction()->getDestinationSiteName();
				mass->eraseOccupiedSite(destinationSite);
				mass->addOccupiedSite(currentSite); // Maybe it can be removed (1)
			}
			(*agent_i)->performAction();
			(*agent_i)->setActionAccepted(true);
		}
	} else {
		for (agent_i = reactingAgents.begin(); agent_i != a_end; ++agent_i) {
			(*agent_i)->setActionAccepted(false);
			(*agent_i)->addDeniedAction(action->createActionInfo());
			(*agent_i)->notifyFailure();
		}
		agent->setActionPerformed(true);
		return false;
	}

	return true;
}

bool MMASS::_manage(spAgent agent, spAction action) {
	spAction nullAction;
	bool ret = true;
                                                                 /*            useless           */
	if ((agent->actionPerformed() && agent->actionAccepted()) || (agent->getAction() != nullAction)) {
		//printf("<font color=\"#006600\">MMASS  - %lu - _manage - <ins>no manage needed</ins> for \"%s\" (<font color=\"#0000ff\">signal</font>)</font><br>\n", turn, agent->getName().c_str()); fflush(NULL);
		return true;
	}

	if (action == nullAction) { // useless
		//printf("<font color=\"#006600\">MMASS  - %lu - _manage - <ins>no action selected</ins> for \"%s\"</font><br>\n", turn, agent->getName().c_str()); fflush(NULL);
		return false;
	}

	ret = action->manage(agent, action);

	//printf("<font color=\"#006600\">MMASS  - %lu - _manage - managing done for \"%s\" (<font color=\"#0000ff\">signal</font>)</font><br>\n", turn, agent->getName().c_str()); fflush(NULL);

	return ret;
}

void* MMASS::_environmentBehaviourThread(void *arg) {
	spAction action;
	spMMASS mmass = CMMASS::getSingleton();
	spAgent_set agents = mmass->getAgentSet();

	FILE *fp = fopen("output.txt", "w+b");
	if (fp == NULL) return NULL;	

	//printf("<font color=\"#006600\">MMASS  - %lu</font><br>\n", mmass->getTurn()); fflush(NULL);
	while (true) {
		spAgent_set agentsChoosingAction = agents;

		while (!agentsChoosingAction.empty()) {
			spAgent_set::iterator agent_i = agentsChoosingAction.begin(),
				a_end = agentsChoosingAction.end();

			while (agent_i != a_end) {
				action = (*agent_i)->collectAction();

				bool actionAccepted = _manage((*agent_i), action);
				action.reset();

				if (actionAccepted) {
					agent_i = agentsChoosingAction.erase(agent_i);
					continue;
				}
			}
			//printf("<font color=\"#006600\">MMASS  - %lu - \"%s\" (turn is %d)</font><br>\n", environmentTurn, (*agent_i)->getName().c_str(), (*agent_i)->getTurn()); fflush(NULL); // este no d pra tirar
			if (agent_i != a_end) ++agent_i;
		}

		_clearFields(mmass);
		_performActions(agents);
		_clearOccupiedSites(mmass);
		//printSitesState(mmass);

		pthread_mutex_lock(&const_cast<pthread_mutex_t>(mmass->getMutex()));
		mmass->nextTurn();
		//printf("MMASS turn is %lu\n", mmass->getTurn()); fflush(NULL);
		while ((mmass->getTurn() > mmass->getOgreTurn()) && (!mmass->finishThread()))
			pthread_cond_wait(&const_cast<pthread_cond_t>(mmass->getCond()), &const_cast<pthread_mutex_t>(mmass->getMutex())); 
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>(mmass->getMutex()));

		//printInfo(agents);
		//printf("<font color=\"#006600\">MMASS  - %lu (<font color=\"#00ff00\">signal</font>)</font><br>\n", mmass->getTurn()); fflush(NULL);

		pthread_mutex_lock(&const_cast<pthread_mutex_t>(mmass->getMutex()));
		if (mmass->finishThread()) {
			pthread_mutex_unlock(&const_cast<pthread_mutex_t>(mmass->getMutex()));
			break;
		}
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>(mmass->getMutex()));
	}

	spAgent_set::const_iterator agent_i, a_end = agents.end();
	for (agent_i = agents.begin(); agent_i != a_end; ++agent_i) {
		(*agent_i)->setFinishThread(true);
	}

	//fprintf(fp, "environment thread finished\n"); fflush(fp);
	fclose(fp);

	return NULL;
}

MMASSDLL_API void MMASS::printInfo(spAgent_set &agents) {
	spAgent_set::const_iterator agent_i, a_end = agents.end();
	for (agent_i = agents.begin(); agent_i != a_end; ++agent_i) {
		if ((*agent_i)->getMASS()->getName() == "mass1")
			if ((*agent_i)->getName() == "agent1.3")
				printf("%s - %s<br>\n", (*agent_i)->getName().c_str(), (*agent_i)->getSite()->getName().c_str()); fflush(NULL);
	}
	printf("<br>\n");
}

MMASSDLL_API void MMASS::printSitesState(spMMASS mmass) {
	set<const string> s;
	spMASS_map::const_iterator mass_i, m_end = mmass->endMASSs();
	for (mass_i = mmass->beginMASSs(); mass_i != m_end; ++mass_i) {
		spSpace space = (*mass_i).second->getSpace();
		spSite_map::const_iterator site_i, s_end = space->endSites();
		for (site_i = space->beginSites(); site_i != s_end; ++site_i) {
			spSite site = (*site_i).second;
			if ((*mass_i).second->getName() == "mass1") {
				printf("%s - FieldAgent = ", site->getName().c_str()); fflush(NULL);
				if (site->fields.count("FieldAgent") > 0) { printf("%7.5g - FieldAgent007 - ", site->fields["FieldAgent"]->getValue()); fflush(NULL); }
				else { printf ("%7.5g - FieldAgent007 - ", (double)0); fflush(NULL); }
				if (site->fields.count("FieldAgent007") > 0) { printf ("%7.5g", site->fields["FieldAgent007"]->getValue()); fflush(NULL); }
				else { printf("%7.5g", (double)0); fflush(NULL); }
				spAgent agent = site->getAgent();
				if (agent) s.insert(agent->getName());
				if (agent) { printf(" - %s", agent->getName().c_str()); fflush(NULL); }
				printf("\n"); fflush(NULL);
			}
		}
	}
	printf("\n"); fflush(NULL);
	if (s.size() != 3) {
		if (s.count("agent1.2") == 0) {
			spSite site = mmass->getMASS("mass1")->getAgent("agent1.2")->getSite();
			printf("agent1.2 is in %s\n", site->getName().c_str()); fflush(NULL);
			spAgent agent = site->getAgent();
			printf("%s has ", site->getName().c_str()); fflush(NULL);
			if (agent) { printf("%s\n", agent->getName().c_str()); fflush(NULL); }
			else { printf("no agent\n"); fflush(NULL); }
		}
		printf("failed! aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa \n");
	}
}

spMMASS CMMASS::self = spMMASS(new CMMASS);