#ifndef MMASS_H
#define MMASS_H

#ifdef MMASSDLL_EXPORTS
#define MMASSDLL_API __declspec(dllexport)
#else
#define MMASSDLL_API __declspec(dllimport)
#endif

#include <pthread.h>
#include "MMASSDefs.h"
#include "MASS.h"
#include <OgreNoMemoryMacros.h>
#include <string>
#include <map>
#include <stdexcept>

namespace MMASS {
	class MMASSDLL_API CMMASS {
	public:
		class logic_error : public std::logic_error {
		public:
			logic_error(const string& msg = "") : std::logic_error(msg) {}
		};

		class invalid_argument : public std::invalid_argument {
		public:
			invalid_argument(const string& msg = "") : std::invalid_argument(msg) {}
		};

	protected:
		spMASS_map masss;
		// interfaces

	public:
		unsigned long LIMIT;
		~CMMASS();

		static const spMMASS getSingleton();

		const spMMASS getSelf() const;

		spAgent_set getAgentSet();

		spMASS createMASS(const string &name_) throw (logic_error, invalid_argument);
		spMASS getMASS(const string &name_) throw (logic_error, invalid_argument);
		const bool hasMASS() const;
		const bool hasMASS(const string &name_) const;
		spMASS_map::const_iterator beginMASSs() const;
		spMASS_map::const_iterator endMASSs() const;
		const spMASS getUnfinishedMASS() const;
		const size_t numOfMASSs() const;

		void finish()  throw (logic_error);
		const bool isReady() const;

		volatile unsigned long getTurn();

		volatile unsigned long getOgreTurn();

		volatile pthread_mutex_t& getMutex();
		volatile pthread_cond_t& getCond();
		volatile pthread_t& getEnvThread();

		void nextOgreTurn();

		volatile bool finishThread();
		void setFinishThread(bool b);

		bool waitOgre();
		void setWaitOgre(bool b);

		void initiate() throw (logic_error);

	private:
		static spMMASS           self;
		bool                     ready;
		unsigned long            turn;
		volatile bool            finish_thread;
		bool                     wait_ogre;
		volatile pthread_mutex_t mutex;
		volatile pthread_cond_t  cond;
		pthread_t       envThread;
		unsigned long            ogreTurn;

		void nextTurn();

		//bool act(spAgent agent, unsigned long turn_);

		CMMASS();

		friend class Agent; // Temporary
		friend class Site;
		friend MMASSDLL_API void* _environmentBehaviourThread(void *arg);
	};

	MMASSDLL_API void _clearOccupiedSites(spMMASS mmass);
	MMASSDLL_API bool _transportManagement(spAgent agent, spAction action);
	MMASSDLL_API bool _emitManagement(spAgent agent, spAction action);
	MMASSDLL_API bool _reactionManagement(spAgent agent, spAction action);
	MMASSDLL_API bool _triggerManagement(spAgent agent, spAction action);
	MMASSDLL_API void _clearFields(spMMASS mmass);
	MMASSDLL_API void _performActions(spAgent_set &agents);
	MMASSDLL_API void printInfo(spAgent_set &agents); // TEMP!!!!!!!!!!!
	MMASSDLL_API void printSitesState(spMMASS mmass); // TEMP!!!!!!!!!!!
}

#endif