#include "Sheep.h"
#include "../MMASSdll/Action.h"
#include "../MMASSdll/Context.h"
#include "../MMASSdll/Site.h"
#include "../MMASSdll/MMASS.h"

namespace shepherdsdog {
	using namespace MMASS;

	Vector3 _getPosition(const string &name) {
		const char *n = name.c_str();
		char num[3];

		int i;
		for (i = 0; n[i] != 'x'; i++) num[i] = n[i];
		num[i++] = '\0';

		int x = atoi(num);

		int j = 0;
		for (; i < name.size(); i++, j++) num[j] = n[i];
		num[j] = '\0';

		int z = atoi(num);

		return Vector3((10 * x) - 5, 5, (10 * z) - 5);
	}

	double FieldMoveAwaySheep::diffusion(spSite source, double generatedValue, spSite destiny) {
		double distance = source->getDistanceOf(destiny);

		if (distance == 0) return generatedValue;
		return (generatedValue / (distance * distance));
	}

	TransportSheep::~TransportSheep() { }

	spAction TransportSheep::createTransportSheep(const string &siteName_) {
		return spAction(new TransportSheep(siteName_));
	}

	void TransportSheep::performAction(spAgent agent) {
		Transport::performAction(agent);
		AgentSheep *sheep = dynamic_cast<AgentSheep*>(agent.get());
		boost::shared_ptr<StateSheep> state = sheep->getState();
		Vector3 v = _getPosition(this->getDestinationSiteName());

		state->addPosition(v);
	}

	TransportSheep::TransportSheep(const string &siteName_) : Transport(siteName_) { }

	bool TransportSheep::manage(spAgent agent, spAction action, unsigned long turn) {
		return MMASS::_transportManagement(agent, action);
	}

	spActionInfo TransportSheep::createActionInfo() {
		return Transport::createActionInfo();
	}


	AgentSheep::~AgentSheep() { fclose(fp); }

	AgentSheep::AgentSheep(const string &name_, boost::shared_ptr<StateSheep> state_)
		:CAgent(name_, state_) {

			fp = fopen("sheepoutput.txt", "w+b");

			this->setCoefficient("FieldMoveAwaySheep", 1);
			this->setThreshold("FieldMoveAwaySheep", 1);

			//this->setCoefficient("FieldAgent007", 2);
			//this->setThreshold("FieldAgent007", 1);{
	}

	void AgentSheep::actionSelect(spContext localContext) {
		spMMASS mmass = CMMASS::getSingleton();

		if ((mmass->getTurn() % 2) == 0) {
			this->setPossibleAction(Emit::createEmit(FieldMoveAwaySheep::newField(5)));
			return;
		} else {
			spSiteInfo site = localContext->getSite();
			spField field;
			spField_set::const_iterator field_i, f_end = site->endFields();
			for (field_i = site->beginFields(); field_i != f_end; ++field_i) {
				if ((*field_i)->getIdentifier() == "FieldMoveAwaySheep") {
					field = (*field_i);
					break;
				}
			}

			if ((field) && (field->getIdentifier() == "FieldMoveAwaySheep")) {
				double fieldValue = field->valuePerceivedForAgent(this->getSelf()), minValue;
				//fprintf(fp, "%g\n", fieldValue); fflush(fp);
				if (fieldValue > 0) {
					minValue = fieldValue;
					string minOwner = site->getSiteName();
					spSiteInfo_set::const_iterator site_i, s_end = localContext->endAdjacentSites();
					for (site_i = localContext->beginAdjacentSites(); site_i != s_end; ++site_i) {
						spField_set::const_iterator f_i, f_e = (*site_i)->endFields();
						for (f_i = (*site_i)->beginFields(); f_i != f_e; ++f_i)
							if ((*f_i)->getIdentifier() == "FieldMoveAwaySheep")
								if ((*f_i)->valuePerceivedForAgent(this->getSelf()) < minValue) {
									minOwner = (*site_i)->getSiteName();
									minValue = (*f_i)->getValue();
								}
					}
					this->setPossibleAction(TransportSheep::createTransportSheep(minOwner));
					return;
				}
			} else {
				this->setPossibleAction(Trigger::createTrigger());
				return;
			}

			/*
			if (!this->hasDeniedAction("TransportSheep")) {
				spSite site = this->getSite();
				Vector3 thisPosition = _getPosition(site->getName());

				wpSite_map::const_iterator site_i, s_end = site->endAdjacentSites();
				for (site_i = site->beginAdjacentSites(); site_i != s_end; ++site_i) {
					Vector3 pos = _getPosition((*site_i).second.lock()->getName());
					if (pos.x > thisPosition.x) break;
				}

				if (site_i == s_end) site_i = site->beginAdjacentSites();

				this->setPossibleAction(TransportSheep::createTransportSheep((*site_i).second.lock()->getName()));
			} else {
				this->setPossibleAction(Trigger::createTrigger());
			}
			*/
		}
		this->setPossibleAction(Trigger::createTrigger());
		return;

		//boost::shared_ptr<StateSheep> state = this->getState();
		//if (state->mWalkList.size() == 1)
			//CMMASS::getSingleton()->setWaitOgre(true);
	}

	const bool AgentSheep::agreeReaction(spAgentInfo_set &involvedAgents) {
		return false;
	}
}