#ifndef SHEEP_H
#define SHEEP_H
#include <pthread.h>
#include "../MMASSdll/MMASSDefs.h"
#include "../MMASSdll/State.h"
#include "../MMASSdll/Agent.h"
#include "../MMASSdll/Action.h"
#include "../MMASSdll/Field.h"
#include <OgreNoMemoryMacros.h>
#include <OgreVector3.h>
#include <string>
#include <deque>

namespace shepherdsdog {
	using namespace MMASS;
	using std::string;
	using std::deque;
	using Ogre::Vector3;

	class FieldMoveAwaySheep : public Field {
	public:
		~FieldMoveAwaySheep() { }

		static spField newField(double value_) { return spField(new FieldMoveAwaySheep(value_)); }

		spField createField(double value_) { return spField(new FieldMoveAwaySheep(value_)); }

		double diffusion(spSite source, double generatedValue, spSite destiny);
		double compose(double value1, double value2) { return (value1 + value2); }
		bool compare(double value1, double value2) { return (value1 >= value2); }

		const string getIdentifier() { return "FieldMoveAwaySheep"; }

	private:
		FieldMoveAwaySheep(double value_) : Field(value_) { }
	};

	class TransportSheep : public Transport {
	public:
		~TransportSheep();

		static spAction createTransportSheep(const string &siteName_);

		void performAction(spAgent agent);

	protected:
		TransportSheep(const string &siteName_);

		const string getIdentifier() { return "Transport"; }

	private:
		bool manage(spAgent agent, spAction action, unsigned long turn);
		spActionInfo createActionInfo();

		friend bool MMASS::_transportManagement(spAgent agent, spAction action);
		friend bool MMASS::_reactionManagement(spAgent agent, spAction action);
	};

	class StateSheep : public State {
	public:
		StateSheep() {
			pthread_mutex_init(&const_cast<pthread_mutex_t>(this->mutex), NULL);
		}
		~StateSheep() {
			pthread_mutex_destroy(&const_cast<pthread_mutex_t>(this->mutex));
		}

		void addPosition(const Vector3 &v) {
			pthread_mutex_lock(&const_cast<pthread_mutex_t>(this->mutex));
			this->mWalkList.push_back(v);
			pthread_mutex_unlock(&const_cast<pthread_mutex_t>(this->mutex));
		}

		const bool empty() {
			pthread_mutex_lock(&const_cast<pthread_mutex_t>(this->mutex));
			bool b = this->mWalkList.empty();
			pthread_mutex_unlock(&const_cast<pthread_mutex_t>(this->mutex));
			return b;
		}
		Vector3 getNext() {
			pthread_mutex_lock(&const_cast<pthread_mutex_t>(this->mutex));
			Vector3 v = this->mWalkList.front();
			this->mWalkList.pop_front();
			pthread_mutex_unlock(&const_cast<pthread_mutex_t>(this->mutex));
			return v;
		}

	protected:
		volatile pthread_mutex_t mutex;
		deque<Vector3> mWalkList;
	};

	class AgentSheep : public CAgent<StateSheep> {
	public:
		FILE *fp;

		class AgentSheepFactory : public Agent::AgentFactory {
			Agent* create(const string &name_) const {
				boost::shared_ptr<StateSheep> state_(new StateSheep);
				return new AgentSheep(name_, state_);
			}
		};

		~AgentSheep();

	protected:
		AgentSheep(const string &name_, boost::shared_ptr<StateSheep> state_);

	private:
		void actionSelect(spContext localContext);
		const bool agreeReaction(spAgentInfo_set &involvedAgents);
		const string getTypeName() { return "AgentSheep"; }

		AgentSheep(const AgentSheep&);
		AgentSheep& operator=(const AgentSheep&);
	};
}
#endif