#ifndef AGENT007_H
#define AGENT007_H
#include "../MMASSdll/MMASSDefs.h"
#include "../MMASSdll/Field.h"
#include "../MMASSdll/State.h"
#include "../MMASSdll/Agent.h"
#include "../MMASSdll/Action.h"
#include "../MMASSdll/Context.h"
#include "../MMASSdll/Site.h"
#include "../MMASSdll/Space.h"
#include "../MMASSdll/MASS.h"
#include "../MMASSdll/MMASS.h"
#include <string>
using namespace MMASS;

class Agent007;

class FieldAgent007 : public Field {
public:
	~FieldAgent007() { }

	static spField newField(double value_) { return spField(new FieldAgent007(value_)); }

	spField createField(double value_) { return spField(new FieldAgent007(value_)); }

	double diffusion(spSite source, double generatedValue, spSite destiny) {
		double distance = source->getDistanceOf(destiny);

		if (distance == 0) return generatedValue;
		return (generatedValue / (distance * distance));
	}

	double compose(double value1, double value2) { return (value1 + value2); }
	bool compare(double value1, double value2) { return (value1 >= value2); }

	const string getIdentifier() { return "FieldAgent007"; }

private:
	FieldAgent007(double value_) : Field(value_) { }
};

class FieldAgent : public Field {
public:
	~FieldAgent() { }

	static spField newField(double value_) { return spField(new FieldAgent(value_)); }

	spField createField(double value_) { return spField(new FieldAgent(value_)); }

	double diffusion(spSite source, double generatedValue, spSite destiny) {
		double distance = source->getDistanceOf(destiny);

		if (distance == 0) return generatedValue;
		return (generatedValue / (distance));
	}

	double compose(double value1, double value2) { return (value1 + value2); }
	bool compare(double value1, double value2) { return (value1 >= value2); }

	const string getIdentifier() { return "FieldAgent"; }

private:
	FieldAgent(double value_) : Field(value_) { }
};

class StateAgent007 : public State {
public:
	unsigned long s;

	StateAgent007() { s = 0; }
};

class ReactionAgent007 : public CReaction<Agent007, StateAgent007> {
public:
	static spAction createReaction(spAgentInfo_set &names) { return spAction(new ReactionAgent007(names)); }

protected:
	ReactionAgent007(spAgentInfo_set &names) : CReaction(names) { }
	const string getIdentifier() { return "Reaction"; }

private:
	void act(spAgent agent, boost::shared_ptr<StateAgent007> state);
};

class Agent007 : public CAgent<StateAgent007> {
public:
	string lastSite;

	class Agent007Factory : public Agent::AgentFactory {
		Agent* create(const std::string &name_) const {
			boost::shared_ptr<StateAgent007> state(new StateAgent007);
			return new Agent007(name_, state);
		}
	};
	~Agent007() { }
	
protected:
	Agent007(const std::string &name_, boost::shared_ptr<StateAgent007> state_) : CAgent(name_, state_) {
		this->setCoefficient("FieldAgent", 1);
		this->setThreshold("FieldAgent", 2);

		this->setCoefficient("FieldAgent007", 2);
		this->setThreshold("FieldAgent007", 1);

		lastSite = "none";
	}

private:
	void actionSelect(spContext localContext) {

		if (this->getName() == "agent1.3") {
			if (this->getSite()->getName() == "site1.2") {
				this->setPossibleAction(Transport::createTransport("site1.7"));
			} else if (this->getSite()->getName() == "site1.7") {
				this->setPossibleAction(Transport::createTransport("site1.8"));
			} else if (this->getSite()->getName() == "site1.8") {
				this->setPossibleAction(Transport::createTransport("site1.9"));
			} else if (this->getSite()->getName() == "site1.9") {
				this->setPossibleAction(Emit::createEmit(FieldAgent::newField(8)));
			}
		} else if (this->getName() == "agent1.1") {
			if (this->getSite()->getName() == "site1.3") {
				this->setPossibleAction(Transport::createTransport("site1.6"));
			} else if (this->getSite()->getName() == "site1.6") {
				this->setPossibleAction(Transport::createTransport("site1.5"));
			} else if (this->getSite()->getName() == "site1.5") {
				this->setPossibleAction(Emit::createEmit(FieldAgent007::newField(6)));
			}
		} else if (this->getName() == "agent1.2") {
			spSiteInfo mySite = localContext->getSite();
			spField_set::const_iterator field_i, f_end = mySite->endFields();
			//printf("%s is in %s\n", this->getName().c_str(), mySite->getSiteName().c_str()); fflush(NULL);
			for (field_i = mySite->beginFields(); field_i != f_end; ++field_i) {
				const double perceived = (*field_i)->valuePerceivedForAgent(this->getSelf());
				//printf("%s perceived %s with %g in %s\n", this->getName().c_str(), (*field_i)->getIdentifier().c_str(), perceived, mySite->getSiteName().c_str()); fflush(NULL);
			}
			//printf("\n"); fflush(NULL);

			if (this->hasDeniedAction("Transport")) {
				this->setPossibleAction(Trigger::createTrigger());
				printf("%s choose trigger!\n\n", this->getName().c_str()); fflush(NULL);
				return;
			}

			if (this->getSite()->getName() == "site1.1") {
				if (lastSite == "site1.2") this->setPossibleAction(Transport::createTransport("site1.3"));
				else this->setPossibleAction(Transport::createTransport("site1.2"));
			} else if (this->getSite()->getName() == "site1.2") {
				if (lastSite == "site1.7") this->setPossibleAction(Transport::createTransport("site1.1"));
				else this->setPossibleAction(Transport::createTransport("site1.7"));
			} else if (this->getSite()->getName() == "site1.3") {
				if (lastSite == "site1.1") this->setPossibleAction(Transport::createTransport("site1.6"));
				else this->setPossibleAction(Transport::createTransport("site1.1"));
			} else if (this->getSite()->getName() == "site1.4") {
				this->setPossibleAction(Transport::createTransport("site1.6"));
			} else if (this->getSite()->getName() == "site1.6") {
				if (lastSite == "site1.3") this->setPossibleAction(Transport::createTransport("site1.4"));
				else this->setPossibleAction(Transport::createTransport("site1.3"));
			} else if (this->getSite()->getName() == "site1.7") {
				if (lastSite == "site1.8") this->setPossibleAction(Transport::createTransport("site1.2"));
				else this->setPossibleAction(Transport::createTransport("site1.8"));
			} else if (this->getSite()->getName() == "site1.8") {
				this->setPossibleAction(Transport::createTransport("site1.7"));
			}

			lastSite = this->getSite()->getName();
		} else {
			this->setPossibleAction(Trigger::createTrigger());
		}
	}

	const bool agreeReaction(spAgentInfo_set &involvedAgents) {
		boost::shared_ptr<StateAgent007> state = this->getState();
		
		if (this->getName() == "agent1.3") {
			if ((this->getTurn() % 2) == 0) return false;
		}

		this->setPossibleAction(ReactionAgent007::createReaction(involvedAgents));
		return true;
	}

	const string getTypeName() { return "Agent007"; }

	//void printState() {
		//boost::shared_ptr<StateAgent007> state_ = this->getState();
		//printf("<font color=\"#000099\">\"%s\" - %lu - %lu actions, %lu actions denied, state is %lu</font><br>\n",  this->getName().c_str(), this->getTurn(), this->action_counter, this->action_denied_counter, state->s); fflush(NULL);
	//}

	Agent007(const Agent007&);
	Agent007& operator=(const Agent007&);
};

#endif