#include "TestMASS.h"
#include "../MMASSdll/Agent.h"
#include "../MMASSdll/Site.h"
#include "../MMASSdll/Space.h"
#include "../MMASSdll/MMASS.h"
#include "../MMASSdll/MMASSDefs.h"
#include "Agent007.h"
using namespace MMASS;

void TestMASS::run() {
	creatingSpaceTests();
}

void TestMASS::creatingSpaceTests() {
	spMMASS mmass = CMMASS::getSingleton();

	// Just to remember how many MASSs
	// were created in the TestMMASS class.
	test_(mmass->numOfMASSs() == 6);

	spMASS mass;
	try {
		mass = mmass->getMASS("mass1");
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Creating a space, from a MASS, with any sites. */
		spSpace space = mass->createSpace();

		test_(space);
		test_(space->getSelf() == space);
		test_(!space->hasSite());
		test_(!space->isReady());
		test_(space->getMASS() == mass);
		test_(mass->getSpace() == space);
		test_(!mass->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create a space from a MASS that already has
		  one space. */
		spSpace space = mass->createSpace();

		fail_("This MASS already has a space."
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Creating some more spaces. */
		mmass->getMASS("mass2")->createSpace();
		mmass->getMASS("mass3")->createSpace();
		mmass->getMASS("mass4")->createSpace();
		mmass->getMASS("mass5")->createSpace();
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}

void TestMASSAgain::run() {
	creatingAgentTests();
	gettingAgentTests();
}

void TestMASSAgain::creatingAgentTests() {
	Agent007::Agent007Factory factory;

	try { /* Creating a agent with no site NOR STATE. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->numOfAgents() == 0);
		test_(mass1->getSpace()->isReady());
		spAgent agent = mass1->createAgent(factory, "agent1.1");

		test_(agent);
		test_(agent->getSelf() == agent);
		// test_(!agent->hasState());
		test_(!agent->hasSite());
		test_(agent->getName() == "agent1.1");
		test_(agent->getMASS() == mass1);
		test_(mass1->hasAgent("agent1.1"));
		test_(mass1->hasAgent(agent));
		test_(mass1->numOfAgents() == 1);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create more than one agent with the same name in the
		  same MASS, even if it is an agent of different type. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->hasAgent("agent1.1"));
		spAgent agent = mass1->createAgent(factory, "agent1.1");

		fail_("This MASS already has a agent with this name."
			"It should have thrown an exception.");
	} catch (MASS::invalid_argument&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->numOfAgents() == 1);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Creating an agent on a site with NO STATE. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->getSpace()->isReady());
		test_(mass1->getSpace()->hasSite("site1.1"));
		spAgent agent = mass1->createAgent(factory, "agent1.2", "site1.1");
		spSite site = mass1->getSpace()->getSite("site1.1");

		test_(agent);
		test_(agent->getSelf() == agent);
		// test_(!agent->hasState());
		test_(agent->hasSite());
		test_(agent->getName() == "agent1.2");
		test_(agent->getMASS() == mass1);
		test_(mass1->hasAgent("agent1.2"));
		test_(mass1->hasAgent(agent));
		test_(mass1->numOfAgents() == 2);
		test_(agent->getSite() == site);
		test_(site->hasAgent());
		test_(site->getAgent() == agent);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create more than one agent with the same name in the
		  same MASS, even if it is an agent of different type. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->hasAgent("agent1.2"));
		spAgent agent = mass1->createAgent(factory, "agent1.2", "site1.1");

		fail_("This MASS already has a agent with this name."
			"It should have thrown an exception.");
	} catch (MASS::invalid_argument&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->numOfAgents() == 2);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create more than one agent on an occupied site. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spSite site = mass1->getSpace()->getSite("site1.1");
		test_(site->hasAgent());
		spAgent agent = mass1->createAgent(factory, "agent1.3", "site1.1");

		fail_("This MASS already has a agent with this name."
			"It should have thrown an exception.");
	} catch (MASS::invalid_argument&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->numOfAgents() == 2);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create an agent with a MASS that have a space
		  that isn't ready (or finished). */
		spMASS mass2 = CMMASS::getSingleton()->getMASS("mass2");
		test_(!mass2->getSpace()->isReady());
		spAgent agent = mass2->createAgent(factory, "agent2.1");

		fail_("This MASS isn't ready."
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass2 = CMMASS::getSingleton()->getMASS("mass2");
		test_(!mass2->hasAgent());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create an agent with a MASS that have a space
		  that isn't ready (or finished). */
		spMASS mass2 = CMMASS::getSingleton()->getMASS("mass2");
		test_(!mass2->getSpace()->isReady());
		test_(mass2->getSpace()->hasSite("site2.1"));
		spAgent agent = mass2->createAgent(factory, "agent2.1", "site2.1");

		fail_("This MASS isn't ready."
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass2 = CMMASS::getSingleton()->getMASS("mass2");
		test_(!mass2->hasAgent());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Creating an agent with NO STATE. */
		spMASS mass3 = CMMASS::getSingleton()->getMASS("mass3");
		test_(mass3->getSpace()->isReady());
		spAgent agent = mass3->createAgent(factory, "agent3.1");

		test_(agent);
		test_(agent->getSelf() == agent);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create more agents than sites in a MASS. */
		spMASS mass3 = CMMASS::getSingleton()->getMASS("mass3");
		test_(mass3->getSpace()->isReady());
		test_(mass3->getSpace()->numOfSites() == 1);
		test_(mass3->numOfAgents() == 1);
		spAgent agent = mass3->createAgent(factory, "agent3.2");

		fail_("This MASS has one site and one agent."
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass3 = CMMASS::getSingleton()->getMASS("mass3");
		test_(mass3->getSpace()->numOfSites() == 1);
		test_(mass3->numOfAgents() == 1);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create more agents than sites in a MASS. */
		spMASS mass3 = CMMASS::getSingleton()->getMASS("mass3");
		test_(mass3->getSpace()->isReady());
		test_(mass3->getSpace()->numOfSites() == 1);
		test_(mass3->numOfAgents() == 1);
		spAgent agent = mass3->createAgent(factory, "agent3.2", "site3.1");

		fail_("This MASS has one site and one agent."
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass3 = CMMASS::getSingleton()->getMASS("mass3");
		test_(mass3->getSpace()->numOfSites() == 1);
		test_(mass3->numOfAgents() == 1);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Creating more agent. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->createAgent(factory, "agent1.3");
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}

void TestMASSAgain::gettingAgentTests() {
	try { /* Can't get a agent from a MASS that doesn't have
		  a space. */
		spMASS mass6 = CMMASS::getSingleton()->getMASS("mass6");
		test_(!mass6->hasSpace());

		mass6->getAgent("smith");

		fail_("This MASS doesn't have a space. "
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass6 = CMMASS::getSingleton()->getMASS("mass6");
		test_(!mass6->hasSpace());
	}catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't get a agent from a MASS that has a space that
		  isn't ready. */
		spMASS mass2 = CMMASS::getSingleton()->getMASS("mass2");
		test_(!mass2->getSpace()->isReady());

		mass2->getAgent("smith");

		fail_("This MASS have a unfinished space . "
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass2 = CMMASS::getSingleton()->getMASS("mass2");
		test_(!mass2->getSpace()->isReady());
	}catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't get a agent from a MASS that doen't have
		  a agent. */
		spMASS mass4 = CMMASS::getSingleton()->getMASS("mass4");
		test_(mass4->getSpace()->isReady());
		test_(!mass4->hasAgent());

		mass4->getAgent("smith");

		fail_("This MASS doen't have a agent. "
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass4 = CMMASS::getSingleton()->getMASS("mass4");
		test_(!mass4->hasAgent());
	}catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't get an inexistent agent. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->getSpace()->isReady());
		test_(mass1->hasAgent());
		test_(!mass1->hasAgent("smith"));

		mass1->getAgent("smith");

		fail_("This MASS doen't have an agent called \"smith\". "
			"It should have thrown an exception.");
	} catch (MASS::invalid_argument&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->hasAgent());
		test_(!mass1->hasAgent("smith"));
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Getting some agents and comparing them. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spMASS mass3 = CMMASS::getSingleton()->getMASS("mass3");

		spAgent agent1, agent3;

		agent1 = mass1->getAgent("agent1.1");
		agent3 = mass3->getAgent("agent3.1");

		test_(agent1);
		test_(agent3);
		test_(agent1 != agent3);
		test_(agent1->getMASS() != agent3->getMASS());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}

void TestMASSFinish::run() {
	buildScenariosForTests();
	finishTests();
}

void TestMASSFinish::buildScenariosForTests() {
	try { /* In mass1, the agent1.3 isn't on a site. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.3");
		spSite site = mass1->getSpace()->getSite("site1.2");
		agent->setSite(site);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}

void TestMASSFinish::finishTests() {
	try { /* In mass1, we have a finished space, we have agents and
		  every agent is on a site. So it can be finished. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(!mass1->isReady());

		mass1->finish();

		test_(mass1->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't finish a MASS that doesn't have a finished space. */
		spMASS mass2 = CMMASS::getSingleton()->getMASS("mass2");
		test_(!mass2->isReady());

		mass2->finish();

		fail_("This MASS doesn't have a finished space. "
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass2 = CMMASS::getSingleton()->getMASS("mass2");
		test_(!mass2->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't finish a MASS that have a agent that isn't on
		  a site. */
		spMASS mass3 = CMMASS::getSingleton()->getMASS("mass3");
		test_(!mass3->isReady());

		mass3->finish();

		fail_("This MASS has a agent that isn't on a site. "
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass3 = CMMASS::getSingleton()->getMASS("mass3");
		test_(!mass3->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't finish a MASS that doesn't have an agent. */
		spMASS mass4 = CMMASS::getSingleton()->getMASS("mass4");
		test_(!mass4->isReady());
		test_(mass4->getSpace()->isReady());
		test_(!mass4->hasAgent());

		mass4->finish();

		fail_("This MASS doesn't have an agent. "
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass4 = CMMASS::getSingleton()->getMASS("mass4");
		test_(!mass4->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't finish a MASS that doesn't have a space. */
		spMASS mass6 = CMMASS::getSingleton()->getMASS("mass6");
		test_(!mass6->isReady());
		test_(!mass6->hasSpace());
		test_(!mass6->hasAgent());

		mass6->finish();

		fail_("This MASS doesn't have a space. "
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass6 = CMMASS::getSingleton()->getMASS("mass6");
		test_(!mass6->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create more agents with a finished MASS. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->numOfAgents() == 3);
		test_(mass1->isReady());
		
		Agent007::Agent007Factory factory;
		spAgent agent = mass1->createAgent(factory, "agent1.4");

		fail_("This MASS is finished. "
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->numOfAgents() == 3);
		test_(mass1->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create more agents with a finished MASS. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->numOfAgents() == 3);
		test_(mass1->isReady());
		
		Agent007::Agent007Factory factory;
		spAgent agent = mass1->createAgent(factory, "agent1.4", "site1.8");

		fail_("This MASS is finished. "
			"It should have thrown an exception.");
	} catch (MASS::logic_error&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		test_(mass1->numOfAgents() == 3);
		test_(mass1->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}
